# note: order in dicts matters (early entries are checked first,
# so better to check new versions first)
class Relion:
    COORD_HEADERS = [f"rlnCoordinate{axis}" for axis in "XYZ"]
    EULER_HEADERS = [f"rlnAngle{angle}" for angle in ["Rot", "Tilt", "Psi"]]
    SHIFT_HEADERS = {
        "4.0": [f"rlnOrigin{axis}Angst" for axis in "XYZ"],
        "3.1": [f"rlnOrigin{axis}Angst" for axis in "XYZ"],
        "3.0": [f"rlnOrigin{axis}" for axis in "XYZ"],
    }
    PIXEL_SIZE_HEADER = {
        "4.0": "rlnTomoTiltSeriesPixelSize",
        "3.1": "rlnImagePixelSize",
        "3.0": "rlnDetectorPixelSize",
    }
    MICROGRAPH_NAME_HEADER = {
        "4.0": "rlnTomoName",
        "3.1": "rlnMicrographName",
        "3.0": "rlnMicrographName",
    }
    OPTICS_GROUP_HEADER = "rlnOpticsGroup"
    POSSIBLE_OPTICS_GROUP_HEADERS = [
        "rlnVoltage",
        "rlnSphericalAberration",
        "rlnAmplitudeContrast",
        "rlnImageSize",
        "rlnImageDimensionality",
        "rlnOpticsGroupName",
        "rlnDetectorPixelSize",
        "rlnImagePixelSize",
        "rlnTomoTiltSeriesPixelSize",
        "rlnMicrographOriginalPixelSize",
        "rlnImageSize",
    ]
    REDUNDANT_HEADERS = (
        COORD_HEADERS
        + EULER_HEADERS
        + SHIFT_HEADERS["3.0"]
        + SHIFT_HEADERS["3.1"]
        + SHIFT_HEADERS["4.0"]
        + [
            PIXEL_SIZE_HEADER["3.0"],
            PIXEL_SIZE_HEADER["3.1"],
            PIXEL_SIZE_HEADER["4.0"],
        ]
    )

    # euler angle definitions are as expected by scipy Rotation objects
    # uppercase is intrinsic, lowercase extrinsic
    EULER = "ZYZ"
    INPLANE = "Z"


class Dynamo:
    COORD_HEADERS = ["x", "y", "z"]
    SHIFT_HEADERS = ["dx", "dy", "dz"]
    EULER_HEADERS = {
        3: ["tdrot", "tilt", "narot"],
        2: ["tilt"],  # TODO: 2d column name might be wrong!
    }
    EXP_ID_HEADER = "tomo"
    EXP_NAME_HEADER = "tomo_file"
    REDUNDANT_HEADERS = COORD_HEADERS + SHIFT_HEADERS + EULER_HEADERS[3]

    EULER = "zxz"
    INPLANE = "z"



import dynamotable

from cryohub.utils.constants import Dynamo
from cryohub.utils.testing import assert_dataframe_equal
from cryohub.writing.tbl import write_tbl

base_columns = [
    *Dynamo.COORD_HEADERS,
    *Dynamo.SHIFT_HEADERS,
    *Dynamo.EULER_HEADERS[3],
    # Dynamo.EXP_ID_HEADER,  # TODO: this currently is lost because we can't write using tomo names
]


def test_write_tbl(tmp_path, poseset, dynamo_tbl):
    file_path = tmp_path / "test.tbl"

    write_tbl(poseset, file_path)
    data = dynamotable.read(file_path)
    assert_dataframe_equal(data, dynamo_tbl, columns=base_columns)



import dask.array as da
import emfile
import numpy as np
from cryotypes.image import Image, validate_image

from ..utils.generic import guess_name


def read_em(image_path, name_regex=None, lazy=True, **kwargs):
    """
    read an em image file
    """
    name = guess_name(image_path, name_regex)

    header, data = emfile.read(image_path, mmap=True)

    if lazy:
        data = da.from_array(data)
    else:
        data = np.asarray(data)

    pixel_size = header["OBJ"] or 0

    tomo = Image(
        data=data,
        experiment_id=name,
        pixel_spacing=pixel_size,
        source=image_path,
        stack=False,
    )

    return validate_image(tomo, coerce=True)



import numpy as np
import pandas as pd
import starfile
from cryotypes.poseset import validate_poseset
from scipy.spatial.transform import Rotation

from ..utils.constants import Relion
from ..utils.generic import ParseError, get_columns_or_default, guess_name
from ..utils.star import merge_optics
from ..utils.types import PoseSet


def construct_poseset(
    df,
    coord_headers,
    shift_headers,
    pixel_size_headers,
    star_path,
    exp_id,
    rescale_shift=False,
    name_regex=None,
):
    coords = get_columns_or_default(df, coord_headers)
    shift = get_columns_or_default(df, shift_headers)
    # relion shift subtract from coordinates
    if shift is not None:
        shift = -shift

    pixel_size = get_columns_or_default(df, pixel_size_headers)
    if pixel_size is not None:
        pixel_size = np.asarray(pixel_size, dtype=float)
        # XXX TODO: remove the following and support variable pixel sizes?
        pixel_size = pixel_size.ravel()[0]
        if rescale_shift and shift is not None:
            shift = shift / pixel_size
    elif rescale_shift:
        raise ParseError(
            "pixel size information is missing, but it is required because shift are in Angstroms!"
        )

    eulers = get_columns_or_default(df, Relion.EULER_HEADERS)
    if eulers is None or np.allclose(eulers, 0):
        rot = None
    else:
        if Relion.EULER_HEADERS[0] in df.columns:
            euler_convention = Relion.EULER
        else:
            euler_convention = Relion.INPLANE
            eulers = eulers[:, 2]
        rot = Rotation.from_euler(euler_convention, eulers, degrees=True)

        # we want the inverse, which when applied to basis vectors it gives us the particle orientation
        rot = rot.inv()

    exp_id = guess_name(exp_id, name_regex)

    features = df.drop(columns=Relion.REDUNDANT_HEADERS, errors="ignore")

    poseset = PoseSet(
        position=coords,
        shift=shift,
        orientation=rot,
        pixel_spacing=pixel_size or 0,
        experiment_id=exp_id,
        source=star_path,
        features=features,
    )

    return validate_poseset(poseset, coerce=True)


def get_proper_header_version(df, header_dict):
    for headers in header_dict.values():
        if isinstance(headers, str):
            if headers in df.columns:
                return headers
        elif any(col in df.columns for col in headers):
            return headers
    return None


def parse_relion_star(
    df,
    star_path="",
    name_regex=None,
    **kwargs,
):
    # find out which columns are present and which version we're dealing with
    rescale_shift = True

    pixel_size_headers = (
        get_proper_header_version(df, Relion.PIXEL_SIZE_HEADER)
        or Relion.PIXEL_SIZE_HEADER["3.0"]
    )
    if pixel_size_headers == Relion.PIXEL_SIZE_HEADER["3.0"]:
        rescale_shift = False

    shift_headers = get_proper_header_version(df, Relion.SHIFT_HEADERS)

    if (
        Relion.COORD_HEADERS[0] not in df.columns
        or Relion.COORD_HEADERS[1] not in df.columns
    ):
        raise ParseError("coordinate information missing")

    coord_headers = Relion.COORD_HEADERS
    exp_header = get_proper_header_version(df, Relion.MICROGRAPH_NAME_HEADER)

    if exp_header is None:
        groups_by_exp = [(None, df)]
    else:
        groups_by_exp = df.groupby(exp_header)

    return [
        construct_poseset(
            exp_df.reset_index(drop=True),
            coord_headers=coord_headers,
            shift_headers=shift_headers,
            pixel_size_headers=pixel_size_headers,
            star_path=star_path,
            exp_id=exp_id,
            rescale_shift=rescale_shift,
            name_regex=name_regex,
        )
        for exp_id, exp_df in groups_by_exp
    ]


def read_star(star_path, **kwargs):
    try:
        raw_data = starfile.read(star_path, always_dict=True)
    except (
        pd.errors.EmptyDataError
    ):  # raised sometimes by .star files with completely different data
        raise ParseError(f"the contents of {star_path} have the wrong format")

    if len(raw_data) == 1:
        df = list(raw_data.values())[0]
    elif "particles" in raw_data and "optics" in raw_data:
        df = merge_optics(raw_data)
    else:
        raise ParseError(
            f"Failed to parse {star_path} as particles. Are you sure this is a particle file?"
        )

    return parse_relion_star(df, star_path=star_path, **kwargs)



import mrcfile
import numpy as np


def _ensure_valid_dtype(data):
    dtype = np.dtype(data.dtype)
    if np.issubdtype(dtype, np.integer):
        # int can only go up to 16bit
        size = min(dtype.itemsize, 2)
        kind = dtype.kind
    elif np.issubdtype(dtype, np.floating):
        # float can only go up to 32bit
        size = min(dtype.itemsize, 4)
        kind = dtype.kind
    else:
        raise TypeError(f'cannot write mrc with dtype "{dtype}"')

    new_dtype = np.dtype(f"{kind}{size}")
    return np.array(data, dtype=new_dtype)


def write_mrc(image, file_path, overwrite=False):
    """
    write an image to disk as an .mrc file
    """
    if not file_path.endswith(".mrc"):
        file_path = file_path + ".mrc"

    mrc = mrcfile.new(file_path, _ensure_valid_dtype(image.data), overwrite=overwrite)
    mrc.set_image_stack() if image.stack else mrc.set_volume()
    mrc.voxel_size = image.pixel_spacing
    mrc.close()



import mrcfile
import starfile

from cryohub.reading.main import read


def test_read_path(tmp_path, image_stack, volume, relion40_star):
    # mrc file
    mrc_path1 = tmp_path / "test1.mrc"
    mrc_path2 = tmp_path / "test2.mrc"
    mrcfile.new(str(mrc_path1), image_stack.data)
    mrcfile.new(str(mrc_path2), volume.data)

    star_path = tmp_path / "test.star"
    starfile.write(relion40_star, star_path)

    data = read(tmp_path, name_regex=r"test\d")

    assert len(data) == 4  # 2 imgs and 2 posesets



import pandas as pd

from .constants import Relion


def extract_optics(df):
    optics_headers = [
        h for h in Relion.POSSIBLE_OPTICS_GROUP_HEADERS if h in df.columns
    ]
    if Relion.OPTICS_GROUP_HEADER in df.columns:
        optic_groups = (
            df.get([Relion.OPTICS_GROUP_HEADER] + optics_headers)
            .drop_duplicates()
            .reset_index(drop=True)
        )
    else:
        if optics_headers:
            df[Relion.OPTICS_GROUP_HEADER] = df.groupby(optics_headers).ngroup()
            optic_groups = (
                df.get([Relion.OPTICS_GROUP_HEADER, *optics_headers])
                .drop_duplicates()
                .reset_index(drop=True)
            )
        else:
            # needed because grouby needs at least a column
            optic_groups = pd.DataFrame({Relion.OPTICS_GROUP_HEADER: [0]})
            df[Relion.OPTICS_GROUP_HEADER] = 0

    df = df.drop(columns=optics_headers, errors="ignore")
    data = {"optics": optic_groups, "particles": df}
    return data


def merge_optics(data_dict):
    return data_dict["particles"].merge(
        data_dict["optics"], on=Relion.OPTICS_GROUP_HEADER
    )



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

import pandas as pd
from numpy.typing import ArrayLike
from scipy.spatial.transform import Rotation


@dataclass
class PoseSet:
    position: ArrayLike
    experiment_id: str
    source: Path | str
    pixel_spacing: float = 0
    shift: ArrayLike | None = None
    orientation: Rotation | None = None
    features: pd.DataFrame | None = None



import h5py

from cryohub.reading.hdf import read_hdf
from cryohub.utils.testing import assert_dataclass_equal

base_fields = ["data", "experiment_id", "pixel_spacing", "stack"]


def test_read_mrc_stack(tmp_path, volume):
    file_path = tmp_path / "test.hdf"

    with h5py.File(file_path, mode="w") as hdf:
        # TODO: how does this vary?
        (
            hdf.create_group("MDF")
            .create_group("images")
            .create_group("0")
            .create_dataset("image", data=volume.data)
        )
    hdf = read_hdf(file_path, name_regex=r"\w+")

    assert_dataclass_equal(hdf, volume, fields=base_fields)



import emfile

from cryohub.writing.em import write_em


def test_write_em(tmp_path, volume):
    file_path = tmp_path / "test.em"
    write_em(volume, str(file_path))
    _, data = emfile.read(file_path)
    assert data.shape == (3, 3, 3)



import starfile

from cryohub.reading.star import read_star
from cryohub.utils.testing import assert_dataclass_equal


def test_read_relion30_3d(tmp_path, relion30_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion30_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    for p, p_exp in zip(part, poseset):
        assert_dataclass_equal(p, p_exp)


def test_read_relion31_3d(tmp_path, relion31_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion31_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    for p, p_exp in zip(part, poseset):
        assert_dataclass_equal(p, p_exp)


def test_read_relion40_3d(tmp_path, relion40_star, poseset):
    file_path = tmp_path / "test.star"
    starfile.write(relion40_star, file_path)

    part = read_star(file_path, name_regex=r"\w")

    for p, p_exp in zip(part, poseset):
        assert_dataclass_equal(p, p_exp)


def test_read_relion40_2d(tmp_path, relion40_star2D, poseset2D):
    file_path = tmp_path / "test.star"
    starfile.write(relion40_star2D, file_path)

    part = read_star(file_path, name_regex=r"\w")

    for p, p_exp in zip(part, poseset2D):
        assert_dataclass_equal(p, p_exp)



import numpy as np
from cryotypes.poseset import validate_poseset

from ..utils.generic import guess_name
from ..utils.types import PoseSet


def read_box(
    path,
    name_regex=None,
    **kwargs,
):
    coords = np.loadtxt(path)
    name = guess_name(path, name_regex)

    poseset = PoseSet(
        position=coords,
        experiment_id=name,
        source=path,
    )
    return validate_poseset(poseset, coerce=True)



from .box import read_box
from .cbox import read_cbox
from .em import read_em
from .hdf import read_hdf
from .main import read
from .mrc import read_mrc
from .star import read_star
from .tbl import read_tbl



import numpy as np
import pandas as pd
import pytest
from cryotypes.image import Image
from scipy.spatial.transform import Rotation

from .utils.types import PoseSet

# some real use case
_angles_relion = np.array(
    [[-18.20820, 138.764387, 10.931106], [-135.12769, 161.921662, -65.26683]]
)

_rotations = Rotation.from_euler("ZYZ", _angles_relion, degrees=True).inv()

_angles_dynamo = _rotations.inv().as_euler("zxz", degrees=True)


@pytest.fixture
def poseset():
    return (
        PoseSet(
            position=np.array([[1, 2, 3]]),
            shift=np.array([[0.01, 0.02, 0.03]]),
            orientation=_rotations[:1],
            experiment_id="a",
            pixel_spacing=10,
            source="test.star",
            features=pd.DataFrame({"feature": ["x"]}),
        ),
        PoseSet(
            position=np.array([[1, 2, 3]]),
            shift=np.array([[0.01, 0.02, 0.03]]),
            orientation=_rotations[1:],
            experiment_id="b",
            pixel_spacing=10,
            source="test.star",
            features=pd.DataFrame({"feature": ["y"]}),
        ),
    )


@pytest.fixture
def poseset2D():
    return (
        PoseSet(
            position=np.array([[1, 2, 0]]),
            shift=np.array([[0.01, 0.02, 0]]),
            orientation=Rotation.from_euler("Z", [60], degrees=True).inv(),
            experiment_id="a",
            pixel_spacing=10,
            source="test.star",
        ),
        PoseSet(
            position=np.array([[1, 2, 0]]),
            shift=np.array([[0.01, 0.02, 0]]),
            orientation=Rotation.from_euler("Z", [45], degrees=True).inv(),
            experiment_id="b",
            pixel_spacing=10,
            source="test.star",
        ),
    )


@pytest.fixture
def relion30_star():
    return pd.DataFrame(
        {
            "rlnCoordinateX": [1, 1],
            "rlnCoordinateY": [2, 2],
            "rlnCoordinateZ": [3, 3],
            "rlnOriginX": [-0.01, -0.01],
            "rlnOriginY": [-0.02, -0.02],
            "rlnOriginZ": [-0.03, -0.03],
            "rlnAngleRot": _angles_relion[:, 0],
            "rlnAngleTilt": _angles_relion[:, 1],
            "rlnAnglePsi": _angles_relion[:, 2],
            "rlnMicrographName": ["a", "b"],
            "rlnDetectorPixelSize": [10, 10],
            "feature": ["x", "y"],
        }
    )


@pytest.fixture
def relion31_star():
    df_optics = pd.DataFrame(
        {
            "rlnOpticsGroup": [1],
            "rlnImagePixelSize": [10],
        }
    )
    df_particles = pd.DataFrame(
        {
            "rlnCoordinateX": [1, 1],
            "rlnCoordinateY": [2, 2],
            "rlnCoordinateZ": [3, 3],
            "rlnOriginXAngst": [-0.1, -0.1],
            "rlnOriginYAngst": [-0.2, -0.2],
            "rlnOriginZAngst": [-0.3, -0.3],
            "rlnAngleRot": _angles_relion[:, 0],
            "rlnAngleTilt": _angles_relion[:, 1],
            "rlnAnglePsi": _angles_relion[:, 2],
            "rlnMicrographName": ["a", "b"],
            "feature": ["x", "y"],
            "rlnOpticsGroup": [1, 1],
        }
    )
    return {"optics": df_optics, "particles": df_particles}


@pytest.fixture
def relion40_star():
    df_optics = pd.DataFrame(
        {
            "rlnOpticsGroup": [1],
            "rlnTomoTiltSeriesPixelSize": [10],
        }
    )
    df_particles = pd.DataFrame(
        {
            "rlnCoordinateX": [1, 1],
            "rlnCoordinateY": [2, 2],
            "rlnCoordinateZ": [3, 3],
            "rlnOriginXAngst": [-0.1, -0.1],
            "rlnOriginYAngst": [-0.2, -0.2],
            "rlnOriginZAngst": [-0.3, -0.3],
            "rlnAngleRot": _angles_relion[:, 0],
            "rlnAngleTilt": _angles_relion[:, 1],
            "rlnAnglePsi": _angles_relion[:, 2],
            "rlnTomoName": ["a", "b"],
            "feature": ["x", "y"],
            "rlnOpticsGroup": [1, 1],
        }
    )
    return {"optics": df_optics, "particles": df_particles}


@pytest.fixture
def relion40_star2D():
    df_optics = pd.DataFrame(
        {
            "rlnOpticsGroup": [1],
            "rlnTomoTiltSeriesPixelSize": [10],
        }
    )
    df_particles = pd.DataFrame(
        {
            "rlnCoordinateX": [1, 1],
            "rlnCoordinateY": [2, 2],
            "rlnOriginXAngst": [-0.1, -0.1],
            "rlnOriginYAngst": [-0.2, -0.2],
            "rlnAnglePsi": [60, 45],
            "rlnTomoName": ["a", "b"],
            "feature": ["x", "y"],
            "rlnOpticsGroup": [1, 1],
        }
    )
    return {"optics": df_optics, "particles": df_particles}


@pytest.fixture
def dynamo_tbl():
    return pd.DataFrame(
        {
            "x": [1, 1],
            "y": [2, 2],
            "z": [3, 3],
            "dx": [0.01, 0.01],
            "dy": [0.02, 0.02],
            "dz": [0.03, 0.03],
            "tdrot": _angles_dynamo[:, 0],
            "tilt": _angles_dynamo[:, 1],
            "narot": _angles_dynamo[:, 2],
            "tomo": [0, 1],
            "feature": ["x", "y"],
        }
    )


@pytest.fixture
def image_stack():
    return Image(
        data=np.ones((3, 3, 3), dtype=np.float32),
        experiment_id="test",
        pixel_spacing=1,
        source="",
        stack=True,
    )


@pytest.fixture
def volume():
    return Image(
        data=np.ones((3, 3, 3), dtype=np.float32),
        experiment_id="test",
        pixel_spacing=1,
        source="",
        stack=False,
    )



import starfile

from cryohub.reading.star import merge_optics
from cryohub.utils.constants import Relion
from cryohub.utils.testing import assert_dataframe_equal
from cryohub.writing.star import write_star

version = "4.0"

base_columns = [
    *Relion.COORD_HEADERS,
    *Relion.SHIFT_HEADERS[version],
    *Relion.EULER_HEADERS,
    Relion.PIXEL_SIZE_HEADER[version],
]


def test_write_star3D(tmp_path, poseset, relion40_star):
    file_path = tmp_path / "test.star"

    write_star(poseset, file_path)
    data = merge_optics(starfile.read(file_path))
    expected = merge_optics(relion40_star)
    assert_dataframe_equal(data, expected, columns=base_columns + ["feature"])


def test_write_star2D(tmp_path, poseset2D, relion40_star2D):
    file_path = tmp_path / "test.star"

    write_star(poseset2D, file_path)
    data = merge_optics(starfile.read(file_path))
    expected = merge_optics(relion40_star2D)
    assert_dataframe_equal(data, expected, columns=base_columns)



import emfile

from cryohub.reading.em import read_em
from cryohub.utils.testing import assert_dataclass_equal

base_fields = ["data", "experiment_id", "pixel_spacing", "stack"]


def test_read_em(tmp_path, volume):
    file_path = tmp_path / "test.em"
    emfile.write(str(file_path), volume.data)
    em = read_em(file_path, name_regex=r"\w+", lazy=False)

    assert_dataclass_equal(em, volume, fields=base_fields)



import re
from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd
from cryotypes.image import ImageProtocol
from cryotypes.poseset import PoseSetProtocol


class ParseError(RuntimeError):
    pass


# a list of commonly used base names for paths in regex form
common_name_regexes = (
    r"^(\w+\d+).*\.",
    r"^(\w+_\d+).*\.",
    r"^(.*\d+).*\.",
    r"^(.*?)\.",
)


def guess_name(string, name_regex=None):
    """
    guess an appropriate name based on the input string
    and a list of regexes in order of priority
    """
    if string is None:
        return "NO_ID"
    string = Path(string).stem
    name_regex = listify(name_regex)
    regexes = list(common_name_regexes)
    regexes = name_regex + regexes
    for regex in regexes:
        if match := re.search(regex, str(string)):
            if match.groups():
                return match.group(1)
            return match.group(0)
    else:
        return "NO_ID"


def pad_to_3D(arr):
    if arr.shape[-1] == 2:
        arr = np.pad(arr, ((0, 0), (0, 1)))
    return arr


def listify(obj):
    """
    transform input into an appropriate list, unless already list-like
    """
    if isinstance(obj, Iterable):
        if isinstance(obj, (str, Path, PoseSetProtocol, ImageProtocol)):
            return [obj]
        else:
            return list(obj)
    if obj is None:
        return []
    return [obj]


def get_columns_or_default(df, columns, default=0):
    """
    Get columns from a dataframe as a numpy array, if present.

    If only some columns are present, fill them with the default value.
    If none are present, return None.
    """
    if columns is None:
        return None
    if isinstance(columns, str):
        columns = [columns]

    cols = {}
    for col in columns:
        cols[col] = df.get(col, default)

    if not any(isinstance(col, pd.Series) for col in cols.values()):
        return None

    return np.asarray(pd.DataFrame(cols, index=df.index))



import mrcfile

from cryohub.reading.mrc import read_mrc
from cryohub.utils.testing import assert_dataclass_equal

base_fields = ["data", "experiment_id", "pixel_spacing", "stack"]


def test_read_mrc_stack(tmp_path, image_stack):
    file_path = tmp_path / "test.mrc"
    mrcfile.new(str(file_path), image_stack.data)
    mrc = read_mrc(file_path, name_regex=r"\w+", lazy=False)

    assert_dataclass_equal(mrc, image_stack, fields=base_fields)


def test_read_mrc_volume(tmp_path, volume):
    file_path = tmp_path / "test.mrc"
    mrcfile.new(str(file_path), volume.data)
    mrc = read_mrc(file_path, name_regex=r"\w+", lazy=False)

    assert_dataclass_equal(mrc, volume, fields=base_fields)



import dask.array as da
import numpy as np
import pandas as pd
from scipy.spatial.transform import Rotation


def rotations_equal(r1, r2):
    """
    Check if rotations result in the same orientations.

    Quaternions result in the same orientation if they have equal real components,
    and imaginary components with the same absolute value
    """
    q1 = r1.as_quat()
    q2 = r2.as_quat()
    if np.allclose(q1[:, :3], q2[:, :3]) and np.allclose(
        np.abs(q1[:, -1]), np.abs(q2[:, -1])
    ):
        return True
    return False


def assert_dataframe_equal(df1, df2, columns=()):
    df1 = df1[[c for c in df1.columns if c in columns]]
    df2 = df2[[c for c in df2.columns if c in columns]]
    assert set(df1.columns) == set(
        df2.columns
    ), f"dataframes have different columns:\n\t{df1.columns}\n\t{df2.columns}"
    for col in df1:
        if isinstance(df1[col][0], Rotation):
            s_rot = Rotation.concatenate(df1[col])
            o_rot = Rotation.concatenate(df2[col])
            # invert and multiply should give zero in mod pi
            if not rotations_equal(s_rot, o_rot):
                raise AssertionError(f'Rotation columns "{col}" of are not equivalent')
        else:
            try:
                if not np.allclose(df1[col], df2[col]):
                    raise AssertionError(
                        f'numeric array columns "{col}" are not within error tolerance'
                    )
            except TypeError:
                if not np.all(df1[col] == df2[col]):
                    raise AssertionError(f'array columns "{col}" are not equal')


def assert_dataclass_equal(data1, data2, fields=()):
    d1 = {
        field: getattr(data1, field)
        for field in data1.__dataclass_fields__
        if field in fields
    }
    d2 = {
        field: getattr(data2, field)
        for field in data2.__dataclass_fields__
        if field in fields
    }
    assert set(d1.keys()) == set(
        d2.keys()
    ), f"models do not have the same set of fields:\n\t{list(d1)}\n\t{list(d2)}"
    for field in d1:
        if isinstance(d1[field], (np.ndarray, da.Array)):
            try:
                if not np.allclose(d1[field], d2[field]):
                    raise AssertionError(
                        f'numeric array fields "{field}" are not within error'
                    )
            except TypeError:
                if not np.all(d1[field] == d2[field]):
                    raise AssertionError(f'array fields "{field}" are not equal')
        elif isinstance(d1[field], pd.DataFrame):
            assert_dataframe_equal(
                d1[field], d2[field], columns=list(d1[field].columns)
            )
        else:
            if not d1[field] == d2[field]:
                return False
    return True



import h5py


def write_hdf(image, file_path, overwrite=False):
    """
    write an image to disk as an .hdf file
    """
    if not file_path.endswith(".hdf"):
        file_path = file_path + ".hdf"

    mode = "w" if overwrite else "x"

    with h5py.File(file_path, mode=mode) as hdf:
        # TODO: how does this vary?
        (
            hdf.create_group("MDF")
            .create_group("images")
            .create_group("0")
            .create_dataset("image", data=image.data)
        )



import dynamotable
import numpy as np
from cryotypes.poseset import validate_poseset
from scipy.spatial.transform import Rotation

from ..utils.constants import Dynamo
from ..utils.generic import get_columns_or_default
from ..utils.types import PoseSet


def read_tbl(table_path, table_map_file=None, name_regex=None, **kwargs):
    """
    Read particles from a dynamo format table file
    """
    df = dynamotable.read(table_path, table_map_file)

    if Dynamo.EXP_NAME_HEADER is df.columns:
        groups_by_exp = [(None, df)]
    else:
        groups_by_exp = df.groupby(Dynamo.EXP_ID_HEADER)

    posesets = []
    for exp_id, exp_df in groups_by_exp:
        exp_df = exp_df.reset_index(drop=True)
        coords = get_columns_or_default(exp_df, Dynamo.COORD_HEADERS)
        shifts = get_columns_or_default(exp_df, Dynamo.SHIFT_HEADERS)
        eulers = get_columns_or_default(exp_df, Dynamo.EULER_HEADERS[3])

        if eulers is None or np.allclose(eulers, 0):
            rot = None
        else:
            if all(header in exp_df.columns for header in Dynamo.EULER_HEADERS[3]):
                euler_convention = Dynamo.EULER
            else:
                euler_convention = Dynamo.INPLANE
                eulers = eulers[:, Dynamo.EULER_HEADERS[2]]
            rot = Rotation.from_euler(euler_convention, eulers, degrees=True)

            # we want the inverse, which when applied to basis vectors it gives us the particle orientation
            rot = rot.inv()

        features = exp_df.drop(columns=Dynamo.REDUNDANT_HEADERS, errors="ignore")

        poseset = PoseSet(
            position=coords,
            shift=shifts,
            orientation=rot,
            experiment_id=exp_id,
            source=table_path,
            features=features,
        )

        posesets.append(validate_poseset(poseset, coerce=True))

    return posesets



import click


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument("paths", nargs=-1)
@click.option(
    "-n",
    "--name-regex",
    metavar="regex",
    help=r"a regex used to infer DataBlock names from paths [fallback: \d+]",
)
@click.option(
    "--strict", is_flag=True, help="immediately fail if a matched path cannot be read"
)
@click.option(
    "--lazy", is_flag=True, default=True, help="read data lazily (if possible)"
)
@click.version_option()
def cli(paths, name_regex, strict, lazy):
    r"""
    cryohub command line interface.

    Opens files and lands in an interactive ipython shell with data loaded as `data`

    PATHS: any number of files [default='.']

    EXAMPLES:

    Open a .star file as particles:

        cryohub particles.star

    Open particles and images from a directory:

        cryohub /dir/with/star_and_mrc_files/

    Match files such as MyProtein_10.star and MyProtein_001.mrc,
    and name the respective data objects Protein_10 and Protein_001:

        cryohub /path/to/dir/MyProtein* -n 'Protein_\d+'
    """
    if not paths:
        paths = ["./*"]

    from IPython.terminal.embed import InteractiveShellEmbed

    import cryohub

    data = cryohub.read(  # noqa: F841
        *paths,
        name_regex=name_regex,
        strict=strict,
        lazy=lazy,
    )

    # set up ipython shell nicely
    banner = "=== cryohub ==="
    sh = InteractiveShellEmbed(banner2=banner)
    sh.push("data")
    sh()



import starfile
from cryotypes.poseset import validate_poseset

from ..utils.generic import guess_name
from ..utils.types import PoseSet


def read_cbox(
    cbox_path,
    name_regex=None,
    **kwargs,
):
    data = starfile.read(cbox_path)["cryolo"]
    coords = data[[f"Coordinate{axis}" for axis in "XYZ"]].to_numpy()
    name = guess_name(cbox_path, name_regex)

    poseset = PoseSet(
        position=coords,
        experiment_id=name,
        source=cbox_path,
    )
    return validate_poseset(poseset, coerce=True)



import mrcfile

from cryohub.writing.mrc import write_mrc


def test_write_mrc_stack(tmp_path, image_stack):
    file_path = tmp_path / "test.mrc"
    write_mrc(image_stack, str(file_path))
    with mrcfile.mmap(str(file_path)) as mrc:
        assert mrc.data.shape == (3, 3, 3)
        assert mrc.voxel_size.item()[0] == 1
        assert mrc.is_image_stack()


def test_write_mrc_volume(tmp_path, volume):
    file_path = tmp_path / "test.mrc"
    write_mrc(volume, str(file_path))
    with mrcfile.mmap(str(file_path)) as mrc:
        assert mrc.data.shape == (3, 3, 3)
        assert mrc.voxel_size.item()[0] == 1
        assert mrc.is_volume()



import dynamotable
import numpy as np
import pandas as pd

from ..utils.constants import Dynamo
from ..utils.generic import listify


def write_tbl(particles, file_path):
    """
    write particle data to disk as a dynamo .tbl file
    """
    particles = listify(particles)

    dataframes = []
    for poseset in particles:
        df = pd.DataFrame()

        if np.allclose(poseset.position[:, 2], 0):
            # 2D data
            df[Dynamo.COORD_HEADERS[:2]] = poseset.position[:, :2]
        else:
            df[Dynamo.COORD_HEADERS] = poseset.position

        shift = poseset.shift
        if shift is not None:
            if np.allclose(shift[:, 2], 0):
                # 2D data
                df[Dynamo.SHIFT_HEADERS[:2]] = shift[:, :2]
            else:
                df[Dynamo.SHIFT_HEADERS] = shift

        # invert rotations for Dynamo and convert to euler (in degrees)
        ori = poseset.orientation.inv()
        if ori is not None:
            rotvec = ori.as_rotvec(degrees=True)
            if np.allclose(rotvec[:, :2], 0):
                # single angle world
                df[Dynamo.EULER_HEADERS[2]] = rotvec[:, 2]
            else:
                df[Dynamo.EULER_HEADERS[3]] = ori.as_euler(Dynamo.EULER, degrees=True)

        # useful to keep around
        df["experiment_id"] = poseset.experiment_id

        # original exp_id source might be saved as features, let's put it back in.
        # dynamo cannot take arbitrary columns, so we're not giving all features
        if poseset.features is not None:
            if Dynamo.EXP_ID_HEADER in poseset.features.columns:
                df[Dynamo.EXP_ID_HEADER] = poseset.features[Dynamo.EXP_ID_HEADER]

        dataframes.append(df)

    df = pd.concat(dataframes)

    if not str(file_path).endswith(".tbl"):
        file_path = str(file_path) + ".tbl"
    dynamotable.write(df, file_path)



import dynamotable

from cryohub.reading.tbl import read_tbl
from cryohub.utils.testing import assert_dataclass_equal


def test_read_tbl(tmp_path, dynamo_tbl, poseset):
    file_path = tmp_path / "test.tbl"
    dynamotable.write(dynamo_tbl, file_path)

    part = read_tbl(file_path, name_regex=r"\w")

    for p, p_exp in zip(part, poseset):
        assert_dataclass_equal(p, p_exp)



import h5py
import numpy as np
from cryotypes.image import Image, validate_image

from ..utils.generic import guess_name


# TODO: lazy mode
def read_hdf(image_path, name_regex=None, **kwargs):
    """
    read an hdf file
    """
    name = guess_name(image_path, name_regex)

    with h5py.File(image_path, mode="r") as hdf:
        # TODO: how does this vary?
        img = hdf["MDF"]["images"]["0"]["image"]
        data = np.asarray(img)

    tomo = Image(
        data=data,
        experiment_id=name,
        pixel_spacing=None or 0,
        source=image_path,
        stack=False,
    )

    return validate_image(tomo, coerce=True)



import numpy as np
import pandas as pd
import starfile

from ..utils.constants import Relion
from ..utils.generic import listify
from ..utils.star import extract_optics


def write_star(particles, file_path, version="4.0", overwrite=False):
    """
    write particle data to disk as a .star file
    """
    particles = listify(particles)

    dataframes = []
    for poseset in particles:
        df = pd.DataFrame()
        if np.allclose(poseset.position[:, 2], 0):
            # 2D data
            df[Relion.COORD_HEADERS[:2]] = poseset.position[:, :2]
        else:
            df[Relion.COORD_HEADERS] = poseset.position

        px_size = poseset.pixel_spacing
        df[Relion.PIXEL_SIZE_HEADER[version]] = px_size

        shift = poseset.shift
        if shift is not None:
            if version != "3.0":
                # shifts are in Angstroms (we need to go to numpy and resize or indices mess up stuff)
                shift = shift * px_size

            # shifts are subtractive in relion
            shift = -shift

            if np.allclose(shift[:, 2], 0):
                # 2D data
                df[Relion.SHIFT_HEADERS[version][:2]] = shift[:, :2]
            else:
                df[Relion.SHIFT_HEADERS[version]] = shift

        # invert rotations for relion and convert to euler (in degrees)
        ori = poseset.orientation.inv()
        if ori is not None:
            rotvec = ori.as_rotvec(degrees=True)
            if np.allclose(rotvec[:, :2], 0):
                # single angle world
                df[Relion.EULER_HEADERS[2]] = rotvec[:, 2]
            else:
                df[Relion.EULER_HEADERS] = ori.as_euler(Relion.EULER, degrees=True)

        # useful to keep around
        df["experiment_id"] = poseset.experiment_id

        if poseset.features is not None:
            df = pd.concat([df, poseset.features.reset_index(drop=True)], axis=1)

        dataframes.append(df)

    df = pd.concat(dataframes)

    # split out optics group if present (and version > 3.0)
    if version != "3.0":
        data = extract_optics(df)
    else:
        data = df

    if not str(file_path).endswith(".star"):
        file_path = str(file_path) + ".star"

    starfile.write(data, file_path, overwrite=overwrite)
